"""
Synth AI - Software for aiding the best and multiplying the will.
"""

from synth_ai.zyk import LM  # Assuming LM is in zyk.py in the same directory

__version__ = "0.1.5"
__all__ = ["LM"]  # Explicitly define public API
